<?php
/**
 * Support functions for evaluating datatypes that accept dates.
 * Support both XSD dates and seconds before/since 1970
 * See http://www.w3.org/TR/xmlschema-2/#dateTime
 * Seconds loses timezone info.
 *
 * @author skierpage
 */

//register type handler:
SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_datetime'),
                       new SMWDateTimeTypeHandler());


/******* Datatype handler classes ********/

class SMWDateTimeTypeHandler {



	/** @public */

	function getID() {
		return 'datetime';
	}

	function getXSDType() {
		return 'http://www.w3.org/2001/XMLSchema#dateTime';
	}

	function getUnits() { //no units for string TODO: maybe seconds since 1970
		return array('STDUNIT'=>'ISO8601', 'ALLUNITS'=> array('ISO8601', 'seconds since 1970') );
	}


    /**
	 * This method transforms the user-provided value of an
	 * attribute into several output strings (one for XML,
	 * one for printout, etc.) and reports parsing errors if
	 * the value is not valid for the given data type.
	 *
	 * @public
	 */
	function processValue($v) {
		$result=Array();
		$result['TYPE']='datetime';	// same as getID();
		$str_val = trim($v);
		if (is_numeric($str_val)) {
			$time = $str_val;
		} else {
			$time = strtotime($str_val);
			if ($time == -1 or $time === false) {
				$result['XSD'] = false;
				$result['USER'] = '<span class="smwwarning">' . wfMsgForContent('smw_nodatetime',$v) . '</span>';
				return $result;
			}
		}
		$result['NUM'] = $time;

		// strtotime accepts non-ISO8601 times like 02/01/70,
		// so  reformat back to ISO8601.
		$str_val = strftime("%Y-%m-%d", $time);
		$user_val = $str_val;
		// See if there is a time component.
		// TODO: what about TimeZone?!
		if ( abs($time - strtotime($str_val)) > 0.5) {
			$str_val .= strftime("T%H:%M:%S", $time);
			$user_val .= strftime(" %H:%M:%S", $time);
		}
		$result['XSD'] = $str_val;
		//$result['USER']= "$user_val (" . SMWNumberFormat($time) . ' seconds since 1970)' ;
		$result['USER']= $user_val; // do not show the seconds since 1970; showing a date in multiple calendar systems could be a future output enhancement (Roman, Gregorian, whatever calendar), if the date is "historical" enough

		// $result['TIP'] = ;


		$result['QUICKSEARCH'] = true;
		//$result['UNIT']='s'; //this should not have a unit string -- it is mostly used for exporting to RDF ...
		return $result;
	}

	/**
	 * This method parses the value in the XSD form that was
	 * generated by parsing some user input. It is needed since
	 * the XSD form must be compatible to XML, and thus does not
	 * respect the internationalization settings. E.g. the German
	 * input value "1,234" is translated into XSD "1.234" which,
	 * if reparsed as a user input would be misinterpreted as 1234.
	 *
	 * @public
	 */
	function processXSDValue($value,$unit) {
		//just insert the local decimal separator
		//(kilo separators do not occur in XSD)
		return $this->processValue(str_replace('.',wfMsgForContent('smw_decseparator'),$value) . $unit);
	}

	/** @private */

	// Could return the numeric seconds since 1970 as well as the timestamp.
	function isNumeric() {
		return TRUE;
	}
}

?>
