<?php
/**
 * This file contains the basic float type and unit conversion methods.
 */

/* Handlers for particular unit conversions. */

SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_float'),
                       new SMWFloatTypeHandler());
SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_geoarea'),
                       new SMWFloatTypeHandler('SMWConvertGeographicArea','geoarea'));
SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_geolength'),
                       new SMWFloatTypeHandler('SMWConvertGeographicLength','geolength'));
SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_mass'),
                       new SMWFloatTypeHandler('SMWConvertMass','mass'));
SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_time'),
                       new SMWFloatTypeHandler('SMWConvertTime','time'));
/**
 * Class for managing floating point types. Parses floating point
 * number strings, supports unit conversion, and generates
 * appropriate error messages on failure.
 *
 * Unit handling is implemented by providing a callback function
 * that takes care of unit conversion. Initializing objects with
 * the empty string as a unit handler allows to use the class as
 * a type handler for plain floating point numbers.
 *
 * @author Markus Krötzsch
 */
class SMWFloatTypeHandler {

	/**#@+
	 * @access private
	 */
	var $uh; //function to call for unit conversion, empty if no unit
	var $typeid; //id string for the datatype object
	/**#@-*/

	function SMWFloatTypeHandler($unitHandler='',$ID='float') {
		$this->uh = $unitHandler;
		$this->typeid = $ID;
	}

	function getID() {
		return $this->typeid;
	}

	function getXSDType() {
		// TODO: Why not #float, or #double ?
		return 'http://www.w3.org/2001/XMLSchema#decimal';
	}

	function getUnits() {
		if ($this->uh!='') {
			$r=call_user_func_array( $this->uh, array(false,false) );
			return array('STDUNIT'=>$r['STDUNIT'], 'ALLUNITS'=>$r['ALLUNITS']);
		} else {
			return array('STDUNIT'=>false, 'ALLUNITS'=>array());
		}
	}

    /**
	 * This method transforms the user-provided value of an
	 * attribute into several output strings (one for XML,
	 * one for printout, etc.) and reports parsing errors if
	 * the value is not valid for the given data type.
	 *
	 * @public
	 */
	function processValue($v) {
		$result=Array();
		$result['TYPE']=$this->typeid;
		// TODO: Also, this isn't ?? multibyte-safe.
		$val = trim($v);
		$decseparator = wfMsgForContent('smw_decseparator');
		$kiloseparator = wfMsgForContent('smw_kiloseparator');

		// Split off number from the rest.
		// Number is, e.g. -12,347,421.55e6
		// Note the separators might be a magic regexp value like '.', so have to escape them with backslash.
		// This rejects .1 , it needs a leading 0.
		// This rejects - 3, there can't be spaces in the number.
		$arr= preg_split('/([-+]?\d+(?:\\' . $kiloseparator . '\d+)*\\' . $decseparator . '?[\d]*(?:\s*[eE]-?\d+)?)[ ]*/', trim($v), 2, PREG_SPLIT_DELIM_CAPTURE);
		// TODO: Having retrieve the number, have to split it off from the rest of the string.  The multiple nested parens make the matcher supply, e.g., 345 as remainder from 12,345.

		// echo 'before=' . $arr[0] . ', number=' . $arr[1] . ', units=' . $arr[2] . '.<br />';
		$num = $arr[1];
		$unitsIn = $arr[2];
		// sscanf doesn't like commas or other than '.' for decimal point.
		$num = str_replace($kiloseparator, '', $num);
		if ($decseparator != '.') {
			$num = str_replace($decseparator, '.', $num);
		}
		// sscanf doesn't like space between number and exponent.
		$num = preg_replace('/\s*([eE]-?\d+)/', '$1', $num, 1);
		// Have to initialize these, a failed sscanf leaves them untouched.
		$float = $extra = '';
		list($float, $extra) = sscanf($num, "%f%s");

		// if stuff before the number, or junk after it, then syntax error
		if (($arr[0]!='') or ($extra!='')) {
			$result['XSD']=false;
			$result['USER']= '<span class="smwwarning">' . wfMsgForContent('smw_nofloat',$v) . '</span>';
			//$result['USER']= '<span class="smwwarning">' . wfMsgForContent('smw_nofloat',$arr[0].'|'.$arr[1].'|'.$arr[2]) . '</span>';
			return $result;
		}

		// TODO: Should this be done on the float or the original number?
		if (mb_substr($num,0,1)=='+') { //strip leading + (relevant for LIKE search)
			$num=substr($num,1);
		}

		$result['QUICKSEARCH'] = true;

		// proceed according to whether units are supported/provided
		if (($this->uh=='') and ($unitsIn=='')) { //no units, just a number
			$result['XSD']=$num; //@TODO: is this fully XSD conformant?  Should I store the float?
			$result['NUM']=$num;
			$result['USER']=SMWNumberFormat($num);
			$result['UNIT']='';
		} else {
			if ($this->uh != '') { //unit handler given
				// So call the handler.
				$units = call_user_func_array( $this->uh, array($num,$unitsIn) );
			} else { //unit given without handler, unsupported unit
				$units=array('ERROR' => wfMsgForContent('smw_unexpectedunit'));
			}

			if (array_key_exists('ERROR',$units)) { //unit not supported by handler
				// proceed anyway, but use given unit instead of data from $units
				$result['XSD']=$num; //@TODO: is this fully XSD conformant?
				$result['NUM']=$num;
				$result['UNIT']=$unitsIn;
				$result['USER']=SMWNumberFormat($num).'&nbsp;'.$unitsIn.' <span class="smwwarning">('.$units['ERROR'].')</span>';
			} else { //unit supported
				// Use the array of values in common units returned by the unit handler.
				$values=$units['VALUES'];

				$i=0;
				$l=count($values);
				$result['USER']='';
				// Iterate over the set of units returned in VALUES array
				// to create UNITS string and possible tooltip.
				// The first unit in VALUES is the primary one displayed in the infobox,
				// the rest appear in parentheses in UNITS
				// Also the first unit's UNIT key is the one returned by this function, so
				// TODO: other UNIT keys set up in the ConvertXxx unit handler are unused?!
				foreach ($values as $unit => $value) {
					// create unit string for infobox
					if ($i==0) {
						$result['XSD']=$value; //@TODO: is this also fully XSD conformant?
						$result['NUM']=$value;
						$result['UNIT']=$unit;
					} elseif ($i==1) {
						$result['USER'].=' (';
					} elseif ($i>1) {
					$result['USER'].=', ';
					}
					$result['USER'].= SMWNumberFormat($value) . '&nbsp;' . $unit;
					if ($i==$l-1) {
						$result['USER'].=')';
					}
					$i++;
					// extend tooltip
					if ($unit!=$units['UNIT']) {
						if ( !array_key_exists('TIP',$result) ) {
							$result['TIP']=$v;
						}
						$result['TIP'].=' = ' . SMWNumberFormat($value) . ' '. $unit;
					}
				}
			}
		}

		return $result;
	}

	/**
	 * This method parses the value in the XSD form that was
	 * generated by parsing some user input. It is needed since
	 * the XSD form must be compatible to XML, and thus does not
	 * respect the internationalization settings. E.g. the German
	 * input value "1,234" is translated into XSD "1.234" which,
	 * if reparsed as a user input would be misinterpreted as 1234.
	 *
	 * @public
	 */
	function processXSDValue($value,$unit) {
		//just insert the local decimal separator
		//(kilo separators do not occur in XSD)
		return $this->processValue(str_replace('.',wfMsgForContent('smw_decseparator'),$value) . $unit);
	}

	function isNumeric() {
		return TRUE;
	}
}



/*************************************************************
 *************************************************************
 *************   Functions for unit conversion   *************
 *************************************************************
 ************************************************************/

/*** Input: $value -- numerical value, $unit -- unit string **
 ***        $value can be FALSE to indicate that only the   **
 ***        mandatory keys are required                     **
 *** Output: array with optional keys                       **
 ***         'ERROR'  error messages if any                 **
 ***         'UNIT'   unit string of originally given unit  **
 ***         'VALUES' array of type "unitstring=>value",    **
 ***                  with first entry being the main unit  **
 ***        and the following *mandatory* keys              **
 ***       'STDUNIT'  Main unit used for storing data, in   **
 ***                  unit column of smw_attributes         **
 ***       'ALLUNITS' An array of supported unit strings    **
 *** All of these require internationalization (errors,     **
 *** unit abbreviations)                                    **/

function SMWConvertGeographicLength($value, $unit) {
	$result=Array();

	$result['STDUNIT']='m';
	$result['ALLUNITS']=array('m', 'km', 'miles');
	if ($value===false) return $result; //abort

	// Input: convert the unit to the main unit
	switch ( $unit ) {
		case '': case 'm': case 'meters': case 'metres':
			$mainval=$value;
			//$result['UNIT']='m';
			break;
		case 'km': case 'kilometers': case 'kilometres':
			$mainval=$value*1000;
			//$result['UNIT']='km';
			break;
		case 'mi': case 'ml': case 'miles': case 'mile':
			$mainval=$value*1609.344;
			//$result['UNIT']='miles';
			break;
		default: //unsupported unit
			$result['ERROR']=wfMsgForContent('smw_unsupportedunit',$unit);
			return $result;
	}

	// Output: create array of the value in common units, the main (STDUNIT) must be first.
	// TODO: attribute should be able to request formats.
	$result['VALUES']=Array();
	$result['VALUES']['m']=$mainval;
	$result['VALUES']['km']=$mainval/1000;
	$result['VALUES']['miles']=$mainval/1609.344;

	return $result;
}

function SMWConvertGeographicArea($value, $unit) {
	$result=Array();

	$result['STDUNIT']='km&sup2;';
	$result['ALLUNITS']=array('km&sup2;', 'ha', 'square kilometres', 'square miles');
	if ($value===false) return $result; //abort

	// Input: convert the unit to the main unit
	if ($unit=='') { // default unit is square kilometers
		$mainval=$value;
	} else {
		switch ( $unit ) {
			case 'ha': case 'hectare': case 'hectares':
				$mainval=$value/100;
				//$result['UNIT']='ha';
				break;
			default: // let's look for squared lengths
				$baseunit=SMWUnSquare($unit);
				switch ($baseunit) {
					case 'km': case 'kilometers': case 'kilometres':
						$mainval=$value;
						//$result['UNIT']='km&sup2;';
						break;
					case 'mi': case 'ml': case 'miles':
						$mainval=$value * 2.589988110336;
						//$result['UNIT']='miles&sup2;';
						break;
					default: //unsupported unit
						$result['ERROR']=wfMsgForContent('smw_unsupportedunit',$unit);
						return $result;
				}
		}
	}
	// Output: create array of the value in common units, the main (STDUNIT) must be first.
	$result['VALUES']=Array();
	$result['VALUES']['km&sup2;']=$mainval;
	$result['VALUES']['ha']=$mainval*100;
	$result['VALUES']['miles&sup2;']=$mainval/2.589988110336;

	return $result;
}


function SMWConvertMass($value, $unit) {
	$result=Array();

	$result['STDUNIT']='kg';	// According to SI, this is the base unit for Mass
	$result['ALLUNITS']=array('g', 'kg', 'tonne', 'lb', 'oz', 'carats', 'u');
	if ($value===false) return $result; //abort

	// Input: convert the unit to the main unit
	switch ( $unit ) {
		case '': case 'kg': case 'kilogram': case 'kilograms':
			$mainval=$value;
			//$result['UNIT']='kg';
			break;
		## TODO: SI prefix handling should be separate, see
		##
		case 'g': case 'grams':
			$mainval=$value/1000;
			//$result['UNIT']='g';
			break;
		## TODO: tonne is the "metric ton", not sure whether to handle imperial ton.
		case 'tonne': case 'tonnes':
			$mainval=$value*1000;
			//$result['UNIT']='tonne';
			break;
		case 'lb': case 'lbs': case 'pounds':
			$mainval=$value*0.45359237; // The official conversion
			//$result['UNIT']='lb';
			break;
		## TODO: would be clearer to use 0.45359237 / 16, but I'm not sure if PHP optimizes that.
		case 'oz':
			$mainval=$value*0.028349523125; // 1/16 of lb conversion
			//$result['UNIT']='oz';
			break;
		case 'carat': case 'carats':
			$mainval=$value*0.2/1000; // 200  milligrams
			//$result['UNIT']='carats';
			break;
		case 'u': case 'Da':
			$mainval=$value*1.66053886*pow(10,-27); // 1.66053886 × 10−27 kg
			//$result['UNIT']='carats';
			break;
		## TODO: imperial tons, etc.
		default: //unsupported unit
			$result['ERROR']=wfMsgForContent('smw_unsupportedunit',$unit);
			return $result;
	}

	// Output: create array of the value in common units, the main (STDUNIT) must be first.
	$result['VALUES']=Array();
	$result['VALUES']['kg']=$mainval;
	$result['VALUES']['g']=$mainval*1000;
	$result['VALUES']['lbs']=$mainval*2.2046226218487758072297380134503; // 1/0.45359237;

	return $result;
}

function SMWConvertTime($value, $unit) {
	$result=Array();

	$result['STDUNIT']='s';	// According to SI, this is the base unit for Mass
	$result['ALLUNITS']=array('s', 'minutes', 'hours', 'days', 'years');
	if ($value===false) return $result; //abort

	// Input: convert the unit to the main unit
	switch ( $unit ) {
		case '': case 's': case 'sec': case 'secs': case 'second': case 'seconds':
			$mainval=$value;
			//$result['UNIT']='s';
			break;
		case 'min': case 'mins': case 'minute': case 'minutes':
			$mainval=$value*60;
			//$result['UNIT']='minutes';
			break;
		case 'hour': case 'hours':
			$mainval=$value*3600;
			//$result['UNIT']='hours';
			break;
		case 'day': case 'days':
			$mainval=$value*86400;
			//$result['UNIT']='days';
			break;
		## TODO: would be clearer to use 0.45359237 / 16, but I'm not sure if PHP optimizes that.
		case 'year': case 'years':
			$mainval=$value*31558149.7676064; // See Wikipedia Year article
			//$result['UNIT']='years';
			break;
		default: //unsupported unit
			$result['ERROR']=wfMsgForContent('smw_unsupportedunit',$unit);
			return $result;
	}

	// Output: create array of the value in common units, the main (STDUNIT) must be first.
	$result['VALUES']=Array();
	$result['VALUES']['s']=$mainval;
	$result['VALUES']['years']=$mainval/31558149.7676064;

	return $result;
}

/**
 * This method checks whether a given input unit string
 * belongs to a squared unit. If this is the case, the
 * unit without the squaring part is returned. Otherwise
 * the function returns FALSE.
 * The purpose of this is to separate the effort of
 * accepting the various html and wiki ways of encoding
 * squaring.
 *
 * TODO: this needs internationalization. Might be not fully trivial,
 *       since the number of ways to encode square is probably not
 *       uniform across languages. Maybe localize the whole method?
 */
function SMWUnSquare($unit) {
	// mb_strlen() is used since strlen() counts each unicode char as 2
	// which is not compatible with substr()
	$len=mb_strlen($unit);

	// The order of checks is chosen so that an early termination is more
	// likely (IMHO). This method might introduce problems for localization.
	if (strpos($unit,'²') === $len-1) { //"²" found at end
		return substr($unit,0,$len-1);
	}
	if (strpos($unit,'&sup2;') === $len-6) { //"&sup2;" found at end
		return substr($unit,0,$len-6);
	}
	if (strpos($unit,'square') ===0) { //"square" found at start
		 //$unit=substr($unit,6);
		 //list(,$unit)=preg_split("/[ ]*/", ' '.$unit, 2); //strip initial spaces
		 //return $unit;
		//probably more efficient:
		return trim(substr($unit,6),' ');
	}
	if (strpos($unit,'sq') ===0) { //"square" found at start
		return substr($unit,2);
	}
	if (strpos($unit,'<sup>2</sup>') === $len-12) { //"<sup>2</sup>" found at end
		return substr($unit,0,$len-12);
	}
	return false;
}

/**
 * Checking for SI prefixes:
 * We need to walk through the units string, peeling off units,
 * attempting to detect SI prefix for each one.
 * E.g. km/hr, kCal, kg·m-1·s-2, etc.
 * Then return the normalized units and a factor.
 * Problems come if the units are in a different order.
 */


?>
