<?php
/**
 * This file contains the datatype management systemsm, some basic
 * classes for process data values given to attributes, and core
 * formatting functions. In order to create new datatypes, copy
 * the simple string or integer type classes below into a separate
 * file, modify as appropriate, add a call to the registration
 * function (SMWRegisterTypeHandler), and include your file here.
 *
 * @author Markus Krötzsch
 * @author Kai Hüner
 */

// initialise arrays *first*
$smwgTypeHandlersByLabel=Array();
$smwgTypeHandlersByID=Array();

require_once('SMW_Storage.php');

/* register datatype handlers */
/**@ TODO: Performance: Note that the code for every type handler is pulled
 *         in, even though most pages do not use all types. 
 *         Better to only load datatype PHP files as needed!?
 */
require_once('SMW_DT_Float.php');
require_once('SMW_DT_GeoCoords.php');
require_once('SMW_DT_DateTime.php');

SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_string'),
                       new SMWStringTypeHandler());
SMWRegisterTypeHandler($smwgContLang->getDatatypeLabel('smw_int'),
                       new SMWIntegerTypeHandler());

	/**
	 * This method registers a typehandler under a certain
	 * label. The label is the name used by the user to denote
	 * this type, without any namespace prefixes.
	 */
	function SMWRegisterTypeHandler($label, $th) {
		global $smwgTypeHandlersByLabel; // typehandler objects indexed by type label
		global $smwgTypeHandlersByID;    // typehandler objects indexed by internal type ID

		$smwgTypeHandlersByLabel[$label] = $th;
		$smwgTypeHandlersByID[$th->getID()] = $th;
	}

	/**
	 * This method returns the XSD datatype identifier
	 * for the type of the given id, or NULL if id is not
	 * known.
	 */
	function SMWGetXSDTypeByID($typeid) {
		global $smwgTypeHandlersByID;    // typehandler objects indexed by internal type ID
		if (array_key_exists($typeid,$smwgTypeHandlersByID)) {
			return $smwgTypeHandlersByID[$typeid]->getXSDType();
		}
		return NULL;
	}

	/**
	 * This method retrieves a type handler object for a given
	 * special property, provided as a constant identifier. Not
	 * every special property has a type (some are relations),
	 * and an error might be returned in certain cases.
	 */
	function &SMWGetSpecialTypeHandler($special) {
		switch ($special) {
			case SMW_SP_HAS_URI:
				return new SMWStringTypeHandler();
			default:
				global $smwgContLang;
				$specprops = $smwgContLang->getSpecialPropertiesArray();
				return new SMWErrorTypeHandler('<span class="smwwarning">' . wfMsgForContent('smw_noattribspecial',$specprops[$special]) . '</span>');
		}
	}

	/**
	 * This method returns the type handler object for a given type label
	 * (i.e. a localized type name), or an error type handler if the label
	 * is not associated with some handler.
	 */
	function &SMWGetTypeHandlerByLabel($typelabel) {
		global $smwgTypeHandlersByLabel;
		if (array_key_exists($typelabel,$smwgTypeHandlersByLabel)) {
			return $smwgTypeHandlersByLabel[$typelabel];
		}
		return new SMWErrorTypeHandler('<span class="smwwarning">' . wfMsgForContent('smw_unknowntype',$typelabel) . '</span>');
	}

	/**
	 * This method retrieves a type handler object for a given
	 * attribute. The attribute is expected to be in text
	 * form without preceding namespace.
	 */
	function &SMWGetTypeHandler($attribute) {
		global $smwgTypeHandlersByAttribute;	// Cache built up by this function.

		if (!is_array($smwgTypeHandlersByAttribute)) {
			$smwgTypeHandlersByAttribute=Array();
		}

		if(!array_key_exists($attribute,$smwgTypeHandlersByAttribute)) {
			global $wgContLang;
			$atitle = Title::newFromText($wgContLang->getNsText(SMW_NS_ATTRIBUTE) . ':' . $attribute);
			if ($atitle->exists()) {
				$typearray = &SMWGetSpecialProperties($atitle,SMW_SP_HAS_TYPE,NULL);
			} else { $typearray = Array(); }
			if (count($typearray)==1) {
				$smwgTypeHandlersByAttribute[$attribute] = SMWGetTypeHandlerByLabel($typearray[0][2]);
			} elseif (count($typearray)==0) {
				$smwgTypeHandlersByAttribute[$attribute] = new SMWErrorTypeHandler('<span class="smwwarning">' . wfMsgForContent('smw_notype') . '</span>');
			} else {
				$smwgTypeHandlersByAttribute[$attribute] = new SMWErrorTypeHandler('<span class="smwwarning">' . wfMsgForContent('smw_manytypes') . '</span>');
			}
		}

		return $smwgTypeHandlersByAttribute[$attribute];
	}

/*
* Basic typehandler classes
* -------------------------
*
* Typehandlers are classes that implement a fixed interface.
*
* The main public interface provided by typehandlers is
* the method processValue($value). It takes the raw string
* representation of a user-provided value and returns an
* array describing the content and possible options for
* display and extended features.
* TODO coming soon: it takes a second value, an array of desired units. 
* The keys in this array
* currently are:
*  USER: the string printed in the infobox, required.
*  XSD : XML Schema representation of single data value,
*        FALSE if value could not be determined, required.
*  NUM : float for representing scalar values, required for
*        all data-types whose values are naturally sortable
*        in a linear way; undefined otherwise.
*  UNIT: unit string if any; normal human-readable text.
*  TIP : tooltip for the value in the article, optional.
*  TYPE: another copy of the type's internal id, required.
*  QUICKSEARCH: if defined, the standard quicksearch link
*               will be printed, optional.
*
*  This interface is subject to change and extension.
*
* A second important function is getUnits() which returns
* an array describing supported units. Available keys are:
*   STDUNIT: the standard unit string as used for storage,
*            or FALSE if no units are supported.
*  ALLUNITS: an array with a wide selection of possible
*            unit strings; typically not all supported string
*            representations, but at least one for every
*            supported unit.
*
* Typehandlers also implement the isNumeric() method, that
* returns TRUE or FALSE, depending on whether the type is
* numeric (i.e. if it saves a value_num into the db)
*/

/**
 * Pseudo typehandler, which returns predefined error
 * messages instead of parsing any value.
 *
 */
class SMWErrorTypeHandler {
	/**#@+
	 * @access private
	 */
	var $emsg; //the message for the user
	/**#@-*/

	/**
	 *  Constructor.
	 */
	function SMWErrorTypeHandler($errorMsg) {
		$this->emsg=$errorMsg;
	}

	function getID() {
		return 'error';
	}

	function getXSDType() {
		return 'http://www.w3.org/2001/XMLSchema#string';
	}

	function getUnits() {
		return array('STDUNIT'=>false, 'ALLUNITS'=>array());
	}

	function processValue($value) {
		$result=Array();
		$result['TYPE']='error';
		$result['XSD']=false;
		$result['USER']=$this->emsg; //send the message
		return $result;
	}

	function processXSDValue($value,$unit) {
		return $this->processValue(NULL);
	}
	
	function isNumeric() {
		return FALSE;
	}
}



/**
 * Class for managing string types. Very simple.
 *
 */
class SMWStringTypeHandler {

	function getID() {
		return 'string';
	}

	function getXSDType() {
		return 'http://www.w3.org/2001/XMLSchema#string';
	}

	function getUnits() { //no units for strings
		return array('STDUNIT'=>false, 'ALLUNITS'=>array());
	}

    /**
	 * This method transforms the user-provided value of an
	 * attribute into several output strings (one for XML,
	 * one for printout, etc.) and reports parsing errors if
	 * the value is not valid for the given data type.
	 *
	 * @public
	 */
	function processValue($value) {
		$result=Array();
		$result['TYPE'] = 'string';
		$result['QUICKSEARCH'] = true;
		if ($value!='') { //do not accept empty strings
			$result['XSD']=$value; //@TODO: one has to check XML-conformance of the string here (replace '<' with '&le;' etc.
			$result['USER']=$value;
			$result['UNIT']='';
		} else {
			$result['XSD']=false;
			$result['USER']='<span class="smwwarning">' . wfMsgForContent('smw_emptystring') . '</span>';
		}
		return $result;
	}

	function processXSDValue($value,$unit) {
		return $this->processValue($value);
	}

	function isNumeric() {
		return FALSE;
	}
}

/**
 * Class for managing integer types. Parses whole number
 * strings and generates appropriate error messages on
 * failure.
 *
 */
class SMWIntegerTypeHandler {

	function getID() {
		return 'int';
	}

	function getXSDType() {
		return 'http://www.w3.org/2001/XMLSchema#integer';
	}

	function getUnits() { //no units for string
		return array('STDUNIT'=>false, 'ALLUNITS'=>array());
	}

    /**
	 * This method transforms the user-provided value of an
	 * attribute into several output strings (one for XML,
	 * one for printout, etc.) and reports parsing errors if
	 * the value is not valid for the given data type.
	 *
	 * @public
	 */
	function processValue($v) {
		$result=Array();
		$result['TYPE']='int';
		// strip kilo separators and split off number from the rest:
		// TODO: this regexp is also in FloatTypeHandler, should move to common.
		$arr= preg_split('/([-+]?[\d]+)/', str_replace(wfMsgForContent('smw_kiloseparator'), '', $v), 2, PREG_SPLIT_DELIM_CAPTURE);

		if (($arr[0]=='') and ($arr[1]!='') and ($arr[2]=='')) { //no junk
			if (mb_substr($arr[1],0,1)=='+') {
				$arr[1]=substr($arr[1],1);  //strip leading + (relevant for LIKE search)
			}
			$result['XSD']=$arr[1];  //@TODO: is this fully XSD conformant?
			$result['NUM']=$arr[1];
			$result['QUICKSEARCH'] = true;
			$result['USER']=SMWNumberFormat($arr[1],0); //"0" -> don't check fractions
			$result['UNIT']='';
		} else {
			$result['XSD']=false;
			$result['USER']='<span class="smwwarning">' . wfMsgForContent('smw_nointeger',$v) . '</span>';
		}

		return $result;
	}

	/**
	 * This method parses the value in the XSD form that was
	 * generated by parsing some user input. It is needed since
	 * the XSD form must be compatible to XML, and thus does not
	 * respect the internationalization settings. E.g. the German
	 * input value "1,234" is translated into XSD "1.234" which,
	 * if reparsed as a user input would be misinterpreted as 1234.
	 *
	 * @public
	 */
	function processXSDValue($value,$unit) {
		return $this->processValue($value . $unit);
	}

	function isNumeric() {
		return TRUE;
	}
}



/**
 * This method formats a float number value according
 * to the given language and precision settings, with
 * some intelligence to produce readable output.
 * Use it whenever you get a number that was not
 * hand-formatted by a user.
 * $decplaces - optional positive integer, controls how many digits after 
 *              the decimal point (but not in scientific notation)
 */
//@TODO: Let this take a format string??!
function SMWNumberFormat($value, $decplaces=3) {
	//@TODO: How to DIE if decplaces is non-integer or negative?
	// : Well, since this function is only called internally, do we have to care much? 
	// : In any case, it would suffice to increase $decplaces to 0 if needed. -- mak
	$decseparator = wfMsgForContent('smw_decseparator');
	$kiloseparator = wfMsgForContent('smw_kiloseparator');
	
	/**
	 * If number is a trillion or more,
	 * then switch to scientific notation.
	 * If number is less than 0.0000001 (i.e. twice decplaces),
	 * then switch to scientific notation.
	 * Otherwise print number using number_format.  This may lead to
	 * 1.200 , so then use rtrim to remove trailing zeroes.
	 */
	$doScientific = false;
	//@TODO: we don't do all this magic for integers, since the formatting does not fit there 
	//       correctly. E.g. one would have integers formatted as 1234e6, not as 1.234e9, right?
	//The "$value!=0" is relevant: we want to scientify numbers that are close to 0, but never 0 ...
	if (($decplaces>0)&&($value!=0)) { 
		if (abs($value) >= 1000000000) {
			$doScientific = true;
		} elseif (abs($value) <= pow(10,-$decplaces)) {
			$doScientific = true;
		} elseif (abs($value) < 1) {
			if (abs($value) <= pow(10,-$decplaces)) {
				$doScientific = true;
			} else {
				// Increase decimal places for small numbers, e.g. .00123 should be 5 places.
				for ($i=0.1; $value <= $i; $i*=0.1) {
					$decplaces++;
				}
			}
		}
	}
	if ($doScientific) {
		// Should we use decimal places here?
		$value = sprintf("%1.6e", $value);
		// Make it more readable by removing trailing zeroes from n.n00e7.
		$value = preg_replace('/(\\.\\d+?)0*e/', '${1}e', $value, 1);
		//NOTE: do not use the optional $count parameter with preg_replace. We need to 
		//      remain compatible with PHP 4.something.
		if ($decseparator !== '.') {
			$value = str_replace('.', $decseparator, $value);
		}
	} else {
		// Format to some level of precision;
		// this does rounding and locale formatting.
		$value = number_format($value, $decplaces, $decseparator, wfMsgForContent('smw_kiloseparator'));
		
		// Make it more readable by removing ending .000 from nnn.000
		//    Assumes substr is faster than a regular expression replacement.
		$end = $decseparator . str_repeat('0', $decplaces);
		$lenEnd = strlen($end);
		if (substr($value, -$lenEnd) === $end ) {
			$value = substr($value, 0, -$lenEnd);
		} else {
			// If above replacement occurred, no need to do the next one.
			// Make it more readable by removing trailing zeroes from nn.n00.
			$value = preg_replace("/(\\$decseparator\\d+?)0*$/", '$1', $value, 1);
		}		
	}
	return $value;
}

?>
