<?php
/**
 * @author Klaus Lassleben
 * @author Markus Krötzsch
 * @author Kai Hüner
 */

	require_once('SMW_Storage.php');
	require_once('SMW_SemanticData.php');

	global $glSemanticData;

	/*******************************************************************************
	H   H  OOO   OOO  K   K  SSS
	H   H O   O O   O K  K  S
	HHHHH O   O O   O KKK    SSS
	H   H O   O O   O K  K      S
	H   H  OOO   OOO  K   K  SSS
	*******************************************************************************/

	/**
	*  This method will be called before an article is displayed or previewed.
	*  For display and preview we strip out the semantic relations and append them
	*  at the end of the article.
	*
	*  TODO: $strip_state is not used (and acutally must not be used, since it is
	*        not relevant when moving the hook to internalParse()).
	*/
	function SMWParserHook(&$parser, &$text, &$strip_state)
	{
		global $glSemanticData;

		smwfInitMessages(); // make sure that the messages are available

		// Global storage for semantic data of this article.
		// The callback functions will store the stripped relations in there.
		$glSemanticData = new SMWSemanticData();

		// Parse links to extract semantic relations
		$semanticLinkPattern = '(\[\[(([^]]*)::)+([^\|\]]*)(\|([^]]*))?\]\])';
		$text = preg_replace_callback($semanticLinkPattern, 'SMWParseRelationsCallback', $text);

		// Parse links to extract attribute values
		$semanticLinkPattern = '(\[\[(([^]]*):=)+([^\|\]]*)(\|([^]]*))?\]\])';
		$text = preg_replace_callback($semanticLinkPattern, 'SMWParseAttributesCallback', $text);

		// print the results

		// Add javascript for tooltip to header
		// Currently, a patch is needed to activate this function, but in future
		// versions this will be included by default
		global $wgOut;
		smwfAddHTMLHeader($wgOut);

		if ( isSemanticsProcessed(&$parser->mTitle) ) {
			$glSemanticData->printInfobox(&$text,&$parser->mTitle,$parser->mOptions->getSkin());
		}
		// just for testing ...
		//$text .= '(Namespace:' . $parser->mTitle->getNamespace() . ")\n";
		//$text .= '(Title:' . $parser->mTitle->getText() . ")\n";
		//global $smwgServer;
		//$text .= '(Server:' . $smwgServer . ")\n";

		return true;
	}


	/**
	 * This hook is triggered latter during parsing to insert HTML code that
	 * would otherwise be escaped by the MediaWiki parser. For example, it is
	 * used to build the JavaScript tooltips.
	 */
	function SMWParserAfterTidyHook(&$parser, &$text)
	{
		// Parse links to extract attribute values
		$semanticLinkPattern = '(<span title="([^"]*)"([^>]*)>([^<]*)</span>)';
		$text = '<div id="toolTip" name="toolTip" class="smwtt"></div>'.preg_replace_callback($semanticLinkPattern, 'SMWParseAttributesAfterTidyCallback', $text);

		return true;
	}

	/**
	*  This method will be called after an article is saved
	*  and stores the semantic relations in the database. One
	*  could consider creating an object for deferred saving
	*  as used in other places of MediaWiki.
	*/
	function SMWSaveHook(&$article, &$user, &$text)
	{
		global $glSemanticData;
		$title=$article->getTitle();
		return $glSemanticData->storeData(&$title, isSemanticsProcessed(&$title));
	}

	/**
	*  This method will be called whenever an article is deleted so that
	*  semantic relations are cleared appropriately.
	*/
	function SMWDeleteHook(&$article, &$user, &$reason)
	{
		$title=$article->getTitle();
		SMWDeleteRelations($title);
		SMWDeleteAttributes($title);
		SMWDeleteSpecialProperties($title);
		return true;
	}

	/**
	*  This method will be called whenever an article is moved so that
	*  semantic relations are moved accordingly.
	*/
	function SMWMoveHook(&$old_title, &$new_title, &$user, $pageid, $redirid)
	{
		SMWMoveAnnotations($old_title, $new_title);
		return true;
	}

	/*******************************************************************************
	 CCC   AAA  L     L     BBBB   AAA   CCC  K   K  SSS
	C     A   A L     L     B   B A   A C     K  K  S
	C     AAAAA L     L     BBBB  AAAAA C     KKK    SSS
	C     A   A L     L     B   B A   A C     K  K      S
	 CCC  A   A LLLLL LLLLL BBBB  A   A  CCC  K   K  SSS
	*******************************************************************************/

	/**
	* This callback function strips out the semantic relation from a wiki link.
	* Semantic relations are stored in the global variable 'glSemanticRelations'
	* for later use.
	*/
	function SMWParseRelationsCallback($semanticLink)
	{
		global $glSemanticData;

		$relation = $semanticLink[2];
		$linkTarget = $semanticLink[3];
		$linkCaption = $semanticLink[4];
		// answer to bug #1479616
		// removes the extra : that comes in from the automatic | expansion in links
		if ($linkCaption[1] == ":") $linkCaption = "|" . substr($linkCaption, 2);

		//extract annotations
		$relations = explode('::', $relation);
		$show_relation_text = FALSE;
		if (':'==$relations[0][0]) {
			$relations[0] = substr($relations[0], 1);
			$show_relation_text = TRUE;
		}
		foreach($relations as $singleRelation) {
			$glSemanticData->addRelation($singleRelation,$linkTarget);
		}

		//pass it back as a normal link
		$result = '[[:' . $linkTarget . $linkCaption . ']]';
		if ($show_relation_text) {
			$result = $relations[0] . ' ' . $result;
		}
		return $result;
	}

	/**
	* This callback function strips out the semantic attributes from a wiki link.
	* Semantic attributes are stored in the global variable 'glSemanticData'
	* for later use.
	*/
	function SMWParseAttributesCallback($semanticLink)
	{
		global $glSemanticData;

		$attribute = $semanticLink[2];
		$value = $semanticLink[3];
		$valueCaption = $semanticLink[4];

		//set text for result
		if ($valueCaption=='') {
			$result = $value;
		}
		else {
			$result = substr( $valueCaption, 1 ); // remove initial '|'
		}

		//extract annotations and create tooltip
		$attributes = explode(':=', $attribute);
		$show_attribute_text = FALSE;
		if ((':'==$attributes[0][0])) {
			$attributes[0] = substr($attributes[0], 1);
			$show_attribute_text = TRUE;
		}
		foreach($attributes as $singleAttribute) {
			$attr = $glSemanticData->addAttribute($singleAttribute,$value);
		}

		//set tooltip for result
		if (array_key_exists('TIP',$attr)) {
			$result = '<span title="'.$attr['TIP'].'" style="color:#B70">'.$result.'</span>';
		}
		if ($show_attribute_text) {
			$result = $attributes[0] . ' ' . $result;
		} //@TODO: this suggests technical improvements

		return $result;
	}


	/**
	 * This callback creates a tooltip based on JavaScript. The creation of
	 * the respective tags has to be performed after "TidyParse" in order to
	 * push the required HTML Tags (a, script) to the code. These would be
	 * escaped when inserted at an earlier stage.
	 */
	function SMWParseAttributesAfterTidyCallback($semanticLink)
	{
		// Here we read the data hosted in the "span"-container. A toolTip
		// is created to show the attribute values. To invoke the tooltip,
		// we have to use a javaScript function (createToolTip) which writes
		// a toolTip (div-Tag) to the document. Since the script-Tag would cause
		// a line break, we have to surround everything with a span-Tag.

		// if the tip is given as an array, we print each value as a line;
		if ((strstr(' = ', $semanticLink[1]) != -1) or (strstr(', ', $semanticLink[1]))) {
			$tip = str_replace(array(' = ', ', '),'<br/>', $semanticLink[1]);
		} else {
			$tip = $semanticLink[1];
		}
		$id      = uniqid('SMW_'); // parameter needed up to PHP4
		$result  = '<span><script type="text/javascript">/*<![CDATA[*/ createToolTip(\''.$id.'\', \''.$tip.'\');/*]]>*/</script>';
		$result .= '<a class="smwatr" onmouseover="showToolTip(\''.$id.'\')" onmouseout="hideToolTip()">'.$semanticLink[3].'</a></span>'; //no CamelCase for onmouse... -> W3C Validator

		return $result;
	}

	/*******************************************************************************
     H E L P E R S
	*******************************************************************************/

	/**
	 * Return true if semantic data should be processed and displayed on this page.
	 * @return bool
	 */
	function isSemanticsProcessed(&$title) {
		global $smwgNamespacesWithSemanticLinks;
		return !empty($smwgNamespacesWithSemanticLinks[$title->getNamespace()]);
	}

?>