<?php
/**
 * The class in this file manages link types, attributes,
 * and other types of semantic data during parsing and
 * storing.
 *
 * @author Markus Krötzsch
 */

require_once('SMW_Datatype.php');
require_once('SMW_Storage.php');

/**
 * Class for representing semantic data, which is
 * in charge of storing and retrieving, but also of
 * parsing the values of attributes provided for
 * articles.
 */
class SMWSemanticData {
	/**#@+
	 * @access private
	 */
	var $attribArray; // Attribute store, elements are arrays indexed by attribute
	                  // names. Each entry in these arrays contains information for
	                  // one value, given as another array with various string
	                  // representations of the value (XSD, human-readable, ...)
	var $relArray;    // Relation store, elements are arrays indexed by relation
	                  // names. Each entry in these arrays contains just another
	                  // array of subjects that were used with this relation.
	var $specaArray;  // Like $attribArray, but for special property attributes,
	                  // indexed by property identifier constants (see SMW_Settings.php)
	var $specrArray;  // Like $relArray, but for special property relations,
	                  // indexed by property identifier constants (see SMW_Settings.php)
	var $typeCache;   // Cache for types of used properties
	//var $articleTitle;// Article to which these attributes belong to
	// >> not needed -- we provide this in a method when storing the data, since it
	// might not be available during creation of this object anyway
	/**#@-*/

	/**
	 * Constructor.
	 */
	function SMWSemanticData() {
		$this->attribArray= Array();
		$this->relArray= Array();
		$this->specaArray= Array();
		$this->specrArray= Array();
		$this->typeCache= Array();
	}

	/**
	 * This method adds a new attribute with the given value to the storage.
	 * It returns an array which contains the result of the operation in various
	 * formats.
	 *
	 * @public
	 */
	function addAttribute($attribute, $value) {
		// See if this attribute is a special one like "Has type"
		global $smwgContLang;
		$attribute = ucfirst($attribute);
		$specprops = $smwgContLang->getSpecialPropertiesArray();
		$special = array_search(ucfirst($attribute), $specprops);

		if ($special!=NULL) { //requires PHP >=4.2.0
			// Initialize attribute storage if needed
			if(!array_key_exists($special,$this->specaArray)) {
				$this->specaArray[$special] = Array();
			}
			$type = &SMWGetSpecialTypeHandler($special);
			//this implements duplicate elimination ... (given that 'USER' is a canonical string)
			$result=$type->processValue($value);
			$result['INPUT'] = $value; //also store the user string
			$this->specaArray[$special][$result['USER'] . $result['XSD'] . $result['UNIT']]=$result;
		} else {
			// Initialize attribute storage if needed
			if(!array_key_exists($attribute,$this->attribArray)) {
				$this->attribArray[$attribute] = Array();
			}
			$type=&SMWGetTypeHandler($attribute);
			  //debug_zval_dump($type); //This was, of course, for debugging ...
			  //It stays here, so I find it when it is needed next time ;-)
			//this implements duplicate elimination ... (given that 'USER' is a canonical string)
			$result=$type->processValue($value);
			$result['INPUT'] = $value; //also store the user string
			$this->attribArray[$attribute][$result['USER'] . $result['XSD'] . $result['UNIT']]=$result;
		}
		return $result;
	}


	/**
	 * This method adds a new relation with the given target to the storage.
	 *
	 * @public
	 */
	function addRelation($relation, $target) {
		global $smwgContLang;
		$relation = ucfirst($relation);
		$srels = $smwgContLang->getSpecialPropertiesArray();
		$special = array_search(ucfirst($relation), $srels);

		if ($special!=NULL) { //requires PHP >=4.2.0
			$type = &SMWGetSpecialTypeHandler($special);
			if ($type->getID() !=  'error') { //Ups! This is not a relation!
				//Note that this still changes the behaviour, since the [[ ]]
				//are not removed! A cleaner solution would be to print a
				//helpful message into the infobox, based on a new "print value as
				//error" datatype handler.
				$this->addAttribute($relation, $target);
			} else {
				// Create a new array for this specific semantic relation
				if(!array_key_exists($special,$this->specrArray)) {
					$this->specrArray[$special] = Array();
				}
				// Store relation and target, if it does not exist yet
				if(!in_array($target, $this->specrArray[$special])) {
					$this->specrArray[$special][] = $target;
				}
			}
		} else {
			// Create a new array for this specific semantic relation
			if(!array_key_exists($relation,$this->relArray)) {
				$this->relArray[$relation] = Array();
			}
			// Store relation and target, if it does not exist yet
			if(!in_array($target, $this->relArray[$relation])) {
				$this->relArray[$relation][] = $target;
			}
		}
		return;
	}

	/*******************************/
	/******* Printing methods ******/
	/*******************************/

	/**
	 * This method prints semantic data at the bottom of an article.
	 *
	 * @public
	 */
	function printInfobox(&$text,&$title,$skin)
	{
		global $wgContLang, $wgServer, $smwgShowInfobox;
		switch ($smwgShowInfobox) {
		case SMW_INFOBOX_HIDDEN: return true;
		case SMW_INFOBOX_NONEMPTY:
			$boxSize = count($this->attribArray) + count($this->relArray) + count($this->specaArray) + count($this->specrArray);
			if ($boxSize == 0) return true;
		}

		// The "\n" is to ensure that lists on the end of articles are terminated
		// before the div starts.
		$text .= "\n" . '<div class="smwinfo">' . "\n";
		$this->printRelations(&$text,&$title,&$skin);
		$this->printAttributes(&$text,&$title,&$skin);
		$this->printSpecialProperties(&$text,&$title,&$skin);
		$text .= '<table style="width: 100%;"><tr>';
		$text .= '<td>';
		$text.=' <p><i>[[' . $wgContLang->getNsText(NS_HELP) . ':' . wfMsgForContent('smw_helppage') . "|''" . wfMsgForContent('smw_edithelp') . "'']]</i></p>\n";
		$text .= '</td>';
		//<td style="width: 50px; vertical-align: bottom; text-align: center; background-color: red;">
		//<span class="rdflink" style="color: blue; width: 200px;">Test</span>
		$text .= '<td style="text-align: right;"><span class="rdflink">';
		$text .= '[' . $wgServer . $skin->makeSpecialUrl('ExportRDF');
		$text .= '/' . str_replace("%2F", "/", urlencode($title->getPrefixedText())) . ' ' . wfMsgForContent('smw_viewasrdf') . ']
		</span></td>';
		$text .= '</tr></table>';
		$text.='</div>';

		//$text.='<div style="background-color: red; float: right;">Test</div>';

		return true;
	}

	/**
	 * This method prints attribute values at the bottom of an article.
	 *
	 * @public
	 */
	function printAttributes(&$text,&$title,&$skin) {
		if(count($this->attribArray) <= 0) { return true; }

		global $wgContLang, $wgServer;
		$text .= ' <p>' . wfMsgForContent('smw_att_head', $title->getText()) . "</p>\n<p>";

		foreach($this->attribArray as $attribute => $attributeValueArray)
		{
			$text .= '   [[' . $wgContLang->getNsText(SMW_NS_ATTRIBUTE). ':' . $attribute . '|' . $attribute . ']]: ';

			$l = count($attributeValueArray);
			$i=0;
			foreach ($attributeValueArray as $attributeValue)
			{
				if ($i!=0) {
					if ($i>$l-2) {
						$text .= wfMsgForContent('smw_finallistconjunct') . ' ';
					} else {
						$text .= ', ';
					}
				}
				$i+=1;
				$text .= $attributeValue['USER'];
				if ($attributeValue['QUICKSEARCH']) {
					$text .= '<!-- -->&nbsp;&nbsp;'; // the comment is neede to prevent MediaWiki from linking URL-strings together with the nbsps!
					$text .= '<span class="smwsearch">[' . $wgServer .
								$skin->makeSpecialUrl('SearchTriple','attribute=' . urlencode($attribute) . '&value=' . urlencode($attributeValue['INPUT']) . '&do=' . urlencode('Search Attributes')) . ' +]</span>';
				}
				if (is_array($attributeValue['INFOLINKS'])) {
					foreach ($attributeValue['INFOLINKS'] as $link) {
						$text .= ' &nbsp;&nbsp;' .$link;
					}
				}
			}
			$text .= "<br/>\n";
		}
		$text .= " </p><hr/>\n";

		return true;
	}

	/**
	 * This method prints semantic relations at the bottom of an article.
	 *
	 * @public
	 */
	function printRelations(&$text,&$title,&$skin) {
		//@ TODO: Performance: remember $NS_RELATION value once, outside this loop;
		//        also in other print loops.
		if(count($this->relArray) <= 0) { return true; }

		global $wgContLang, $wgServer;
		$text .= ' <p>' . wfMsgForContent('smw_rel_head') . "</p>\n<p>";

		foreach($this->relArray as $relation => $relationObjectArray)
		{
			$text .= '   ' . $title->getPrefixedText();
			$text .= '&nbsp; [[' . $wgContLang->getNsText(SMW_NS_RELATION). ':' . $relation . '|' . $relation . ']]&nbsp; ';

			$l = count($relationObjectArray);
			for ($i=0; $i<$l; $i++)
			{
				if ($i!=0) {
					if ($i>$l-2) {
						$text .= wfMsgForContent('smw_finallistconjunct') . ' ';
					} else {
						$text .= ', ';
					}
				}

				$relationObject = $relationObjectArray[$i];

				//$t=Title::newFromText($relationObject);
				$text .= '[[:' . $relationObject . ']]';
				$text .= '&nbsp;&nbsp;<span class="smwsearch">[' . $wgServer . $skin->makeSpecialUrl('SearchTriple', 'relation=' . urlencode($relation) . '&object=' . urlencode($relationObject) . '&do=' . urlencode('Search Relations'));
				$text .= ' +]</span>';
			}
			$text .= "<br/>\n";
		}
		$text .= " </p><hr/>\n";

		return true;
	}


	/**
	 * This method prints special properties at the bottom of an article.
	 *
	 * @public
	 */
	function printSpecialProperties(&$text,&$title,&$skin) {
		global $wgContLang, $smwgContLang;

		if ((count($this->specaArray) > 0)||(count($this->specrArray) > 0)) {
			$text .= ' <p>' . wfMsgForContent('smw_spec_head', $title->getText()) . "</p>\n<p>";
		} else { return true; }

		$specprops = $smwgContLang->getSpecialPropertiesArray();
		$titletext = $title->getPrefixedText();

		if(count($this->specaArray) > 0) {
			foreach($this->specaArray as $specialProperty => $valueArray) {
				$specialPropertyName = $specprops[$specialProperty];
				foreach ($valueArray as $value) {
					$text .= '   ' . $titletext;
					$text .= '&nbsp; [[' . $wgContLang->getNsText(SMW_NS_ATTRIBUTE). ':' . $specialPropertyName . '|' . $specialPropertyName . ']]&nbsp; ';
					$text .= $value['USER'];
					$text .= "<br/>\n";
				}
			}
		}

		if(count($this->specrArray) > 0) {
			foreach($this->specrArray as $specialProperty => $valueArray)
			{
				$specialPropertyName = $specprops[$specialProperty];
				foreach ($valueArray as $value) {
					$text .= '   ' . $titletext;
					$text .= '&nbsp; [[' . $wgContLang->getNsText(SMW_NS_RELATION). ':' . $specialPropertyName . '|' . $specialPropertyName . ']]&nbsp; ';
					$text .= '[[' . $value . ']]';
					$text .= "<br/>\n";
				}

			}
		}

		$text .= " </p><hr/>\n";

		return true;
	}

	/*******************************/
	/******* Storage methods *******/
	/*******************************/

	/**
	 * This method stores the semantic data.
	 *
	 * @public
	 */
	function storeData(&$title, $processSemantics) {
		// clear data even if semantics are not processed for this namespace
		// (this setting might have been changed, so that data still exists)
		$this->clearData($title);
		if ($processSemantics) {
			$this->storeAttributes($title);
			$this->storeRelations($title);
			$this->storeSpecialProperties($title);
		}
		return true;
	}

	/**
	 * Method for deleting semantic data currently associated
	 * with some article.
	 */
	function clearData($s_title) {
		SMWDeleteRelations($s_title);
		SMWDeleteAttributes($s_title);
		SMWDeleteSpecialProperties($s_title);
	}

	/**
	 * Method for storing attributes.
	 *
	 * @public
	 */
	function storeAttributes($s_title) {
		if(count($this->attribArray) <= 0) {
			return true;
		}

		foreach($this->attribArray as $attribute => $attributeValueArray) {
			$a_title = Title::newFromText($attribute,SMW_NS_ATTRIBUTE);
			if ($a_title !== NULL) {
				foreach($attributeValueArray as $value) {
					if ($value['XSD']!==false) {
						SMWStoreAttribute($s_title, $a_title, $value['UNIT'], $value['TYPE'], $value['XSD'], $value['NUM']);
					}
				}
			}
		}
		return true;
	}

	/**
	 * Method for storing semantic relations.
	 *
	 * @public
	 */
	function storeRelations($s_title) {
		if(count($this->relArray) <= 0) {
			return true;
		}

		foreach($this->relArray as $semanticRelation => $relationObjectArray)
		{
			$r_title = Title::newFromText($semanticRelation,SMW_NS_RELATION);
			if ($r_title !== NULL) {
				foreach($relationObjectArray as $linkTarget)
				{
					$o_title = Title::newFromText($linkTarget);
					if ( $o_title !== NULL ) {
						SMWStoreRelation($s_title, $r_title, $o_title);
					}
				}
			}
		}
		return true;
	}

	/**
	 * Method for storing special properties.
	 *
	 * @public
	 */
	function storeSpecialProperties($s_title) {
		if (count($this->specrArray) > 0) {
			foreach($this->specrArray as $specialProperty => $valueArray)
			{
				foreach($valueArray as $value)
				{
					SMWStoreSpecialProperty($s_title, $specialProperty, $value);
				}
			}
		}

		if (count($this->specaArray) > 0) {
			foreach($this->specaArray as $specialProperty => $valueArray) {
				foreach($valueArray as $value) {
					if ($value['XSD']!==false) {
						SMWStoreSpecialProperty($s_title, $specialProperty, $value['XSD']);
					}
				}
			}
		}
		return true;
	}

}

?>
